# -*- coding: utf-8 -*-
"""pdns module"""

__license__ = """
    Copyright (C) 2018  fjord-technologies

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along
    with this program; if not, write to the Free Software Foundation, Inc.,
    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA..
"""

import json
import logging
import requests
import time
import uuid

from dwho.classes.modules import DWhoModuleBase, MODULES
from nsaproxy.classes.apis import NSAProxyApiObject, APIS_SYNC
from sonicprobe.libs import network, urisup, xys
from sonicprobe.libs.moresynchro import RWLock
from sonicprobe.libs.http_json_server import HttpReqErrJson, HttpResponseJson


LOG     = logging.getLogger('nsaproxy.modules.pdns')


def validate_params(params):
    if not isinstance(params, (list, dict)):
        return False
    elif len(params) == 0:
        return True

    return params

def validate_name(ip_addr):
    return network.valid_host(ip_addr, network.MASK_IPV4_DOTDEC | network.MASK_IPV6)


xys.add_callback('pdns.validate_params', validate_params)
xys.add_callback('pdns.validate_nameservers', network.valid_host)


class PDNSModule(DWhoModuleBase):
    MODULE_NAME     = 'pdns'

    LOCK            = RWLock()

    def safe_init(self, options):
        self.results      = {}
        self.lock_timeout = self.config['general']['lock_timeout']
        self.pdns_uri     = list(urisup.uri_help_split(self.config['general']['pdns_uri']))

    @staticmethod
    def _build_http_response(data = ''):
        req = HttpResponseJson(data = data)
        req.add_header('Access-Control-Allow-Origin', '*')
        req.add_header('Access-Control-Allow-Methods', 'OPTIONS, POST')
        req.add_header('Access-Control-Allow-Headers', 'Origin, X-Requested-With, Content-Type, Accept, Authorization')
        req.add_header('Access-Control-Max-Age', 1728000)

        return req

    @staticmethod
    def _mk_hash(salt, val):
        return hashlib.sha256("%s:%s" % (salt, val)).hexdigest()

    def _dashboard_params(self, request, args):
        if 'dashboard' not in self.config:
            return args

        for dashboard, params in self.config['dashboard'].iteritems():
            if self.path_referer.endswith("/%s" % dashboard):
                return helpers.merge(args, params)

        return args

    def _do_request(self, method, path, payload, headers):
        uri    = list(self.pdns_uri)
        if path:
            uri[2] = path
        else:
            uri[2] = None

        return getattr(requests, method.lower())(urisup.uri_help_unsplit(uri),
                                                 json    = payload,
                                                 headers = headers)

    def _set_auth_arg(self, args):
        if 'auth' not in args:
            return args

        auth = self.adapter_redis.get_key(self.ga_key)
        if auth:
            args['auth'] = auth

        return args

    def _crypt_groups(self, result, groups):
        r = []

        if not isinstance(result, list):
            return r

        for group in result:
            if group['groupid'] in groups:
                grouphash        = self._mk_hash(self.salt, group['groupid'])
                self.adapter_redis.set_key("groupid:%s" % grouphash, group['groupid']) 
                group['groupid'] = grouphash
                r.append(group)

        return r

    def _crypt_hosts(self, result, hosts):
        r = []

        if not isinstance(result, list):
            return r

        for host in result:
            if 'all' in hosts or host['hostid'] in hosts:
                hosthash       = self._mk_hash(self.salt, host['hostid'])
                self.adapter_redis.set_key("hostid:%s" % hosthash, host['hostid'])
                host['hostid'] = hosthash
                r.append(host)

        return r

    def _crypt_ids(self, key, result):
        r = []

        if not isinstance(result, list):
            return r

        for value in result:
            xhash = self._mk_hash(self.salt, value[key])
            self.adapter_redis.set_key("%s:%s" % (key, xhash), value[key]) 
            value[key] = xhash
            r.append(value)

        return r

    def _decrypt_ids(self, key, ids):
        r = []

        if not isinstance(ids, list):
            return r

        for xhash in ids:
            xid = self.adapter_redis.get_key("%s:%s" % (key, xhash))
            if not xid:
                raise HttpReqErrJson(400, "invalid %s" % key)
            r.append(xid)

        return r

    def _ck_crypt_ids(self, xdict):
        if not isinstance(xdict, dict):
            return xdict

        for key in KEYS_ID:
            if key not in xdict:
                continue
            xhash = self._mk_hash(self.salt, xdict[key])
            self.adapter_redis.set_key("%s:%s" % (key, xhash), xdict[key])
            xdict[key] = xhash

        return xdict

    def _ck_decrypt_ids(self, xdict):
        if not isinstance(xdict, dict):
            return xdict

        for key in KEYS_ID:
            if key in xdict:
                xdict[key] = self.adapter_redis.get_key("%s:%s" % key, xdict[key])

        return xdict

    @staticmethod
    def _return_no_result(args):
        return {'jsonrpc': args['jsonrpc'],
                'id':      args['id'],
                'result':  []}

    def _method_default(self, args, request):
        return self._do_request(json.dumps(args), request.headers.dict)

    def _method_apiinfo_version(self, args, request):
        return self._do_request(json.dumps(args), request.headers.dict)

    def _method_user_login(self, args, request):
        guser = self.objects['user'].get(**{'login': self.ga_user})
        if not guser:
            raise HttpReqErrJson(404, "grafana user not found")

        if not self.objects['access'].exists(**{'user_id': guser['id']}):
            raise HttpReqErrJson(404, "user not found")

        payload = json.dumps({'jsonrpc': args['jsonrpc'],
                              'method':  args['method'],
                              'params':  {'user':     self.pdns_user,
                                          'password': self.pdns_passwd},
                              'id':      args['id']})

        r = self._do_request(payload, request.headers.dict)

        if not r or 'result' not in r:
            raise HttpReqErrJson(401, "unauthorized")

        self.adapter_redis.set_key(self.ga_key,
                                   r['result'],
                                   self.key_expire)
        self.adapter_redis.set_key("guid:%s" % self.ga_user,
                                   guser['id'],
                                   self.key_expire)
        self.adapter_redis.set_key("salt:%s" % self.ga_key,
                                   ("%s" % uuid.uuid4()).replace('-', ''),
                                   self.key_expire)

        r['result'] = self.ga_key 

        return r

    def _method_hostgroup_get(self, args, request):
        res = self.objects['access'].get_all(
                  **{'user_id': self.adapter_redis.get_key("guid:%s" % self.ga_user)})
        if not res:
            self._return_no_result(args)

        groups = set()

        for row in res:
            groups.add(str(row['hostgroup_id']))

        r = self._do_request(json.dumps(args), request.headers.dict)
        if not r or 'result' not in r:
            raise HttpReqErrJson(404, "no such hostgroup")

        r['result'] = self._crypt_groups(r['result'], groups)

        return r

    def _method_host_get(self, args, request):
        if not self.ids['groupids']:
            raise HttpReqErrJson(400, "missing groupids")

        res = self.objects['access'].get_all(
                  **{'user_id':      self.adapter_redis.get_key("guid:%s" % self.ga_user),
                     'hostgroup_id': self.ids['groupids']})
        if not res:
            self._return_no_result(args)

        hosts = set()

        for row in res:
            if row['host_id'] is not None:
                hosts.add(str(row['host_id']))
            else:
                hosts.add('all')
                break

        args['params']['groupids'] = self.ids['groupids']

        r = self._do_request(json.dumps(self._dashboard_params(request, args)),
                             request.headers.dict)
        if not r or 'result' not in r:
            raise HttpReqErrJson(400, "no such host")

        r['result'] = self._crypt_hosts(r['result'], hosts)

        return r

    def _method_application_get(self, args, request):
        if not self.ids['hostids']:
            self._return_no_result(args)

        args['params']['hostids'] = self.ids['hostids']

        r = self._do_request(json.dumps(args), request.headers.dict)
        if not r or 'result' not in r:
            raise HttpReqErrJson(400, "no such application")

        r['result'] = self._crypt_ids('applicationid', r['result'])

        return r

    def _method_trigger_get(self, args, request):
        if not self.ids['hostids']:
            raise HttpReqErrJson(400, "missing hostids")

        if 'groupids' in args['params']:
            args['params']['groupids'] = self.ids['groupids']

        args['params']['hostids'] = self.ids['hostids']

        r = self._do_request(json.dumps(args), request.headers.dict)
        if not r or 'result' not in r:
            raise HttpReqErrJson(400, "no such trigger")

        for x in r['result']:
            for k in ('groups', 'hosts', 'items'):
                if k not in x:
                    continue
                x[k] = map(self._ck_crypt_ids, x[k])
            x = self._ck_crypt_ids(x)

        return r

    def _method_item_get(self, args, request):
        if 'applicationids' in args['params']:
            args['params']['applicationids'] = self.ids['applicationids']

        r = self._do_request(json.dumps(args), request.headers.dict)
        if not r or 'result' not in r:
            raise HttpReqErrJson(400, "no such item")

        for x in r['result']:
            if 'hosts' in x:
                x['hosts'] = map(self._ck_crypt_ids, x['hosts'])
            x = self._ck_crypt_ids(x)

        return r

    def _method_event_get(self, args, request):
        if 'objectids' in args['params']:
            args['params']['objectids'] = self.ids['objectids']

        r = self._do_request(json.dumps(args), request.headers.dict)
        if not r or 'result' not in r:
            raise HttpReqErrJson(400, "no such event")

        for x in r['result']:
            x = self._ck_crypt_ids(x)

        return r

    def _method_history_get(self, args, request):
        if not self.ids['itemids']:
            raise HttpReqErrJson(400, "missing itemids")

        args['params']['itemids'] = self.ids['itemids']

        r = self._do_request(json.dumps(args), request.headers.dict)
        if not r or 'result' not in r:
            raise HttpReqErrJson(400, "no such history")

        for x in r['result']:
            x = self._ck_crypt_ids(x)

        return r

    def _method_trend_get(self, args, request):
        if not self.ids['itemids']:
            raise HttpReqErrJson(400, "missing itemids")

        args['params']['itemids'] = self.ids['itemids']

        r = self._do_request(json.dumps(args), request.headers.dict)
        if not r or 'result' not in r:
            raise HttpReqErrJson(400, "no such trend")

        for x in r['result']:
            x = self._ck_crypt_ids(x)

        return r

    def _method_service_get(self, args, request):
        return self._return_no_result(args)

    def _set_result(self, obj):
        self.results[obj.get_uid()] = obj

    def _get_results(self, uids):
        r = {'expected':   len(uids),
             'successful': [],
             'failed':     []}

        while r['expected'] > 0:
            for uid in uids:
                if uid not in self.results:
                    time.sleep(0.5)
                    continue

                res = self.results.pop(uid)
                if res.has_error():
                    r['failed'].append(res)
                    r['expected'] -= 1
                    LOG.warning("failed on call: %r. (errors: %r)", res.get_uid(), res.get_errors())
                else:
                    r['successful'].append(res)
                    r['expected']   -= 1
                    LOG.info("successful on call: %r. (result: %r)", res.get_uid(), res.get_result())

        return r

    def _push_apis_sync(self, endpoint, params, args = None):
        r   = []
        xid = "%s" % uuid.uuid4()
        for api_sync in APIS_SYNC.itervalues():
            uid = "%s:%s" % (api_sync.name, xid)
            api_sync.qput(NSAProxyApiObject(uid, endpoint, params, args, self._set_result))
            r.append(uid)

        return r


    ENDPOINT_GET_QSCHEMA = xys.load("""
    server_id: !!str
    endpoint:  !!str
    id*:       !!str
    """)

    def api_endpoint_get(self, request):
        params  = request.query_params()

        api_key = request.headers.get('x-api-key')
        if not api_key:
            raise HttpReqErrJson("missing api key")

        if not isinstance(params, dict):
            raise HttpReqErrJson(400, "invalid arguments type")

        if not xys.validate(params, self.ENDPOINT_GET_QSCHEMA):
            raise HttpReqErrJson(415, "invalid arguments for command")

        if not self.LOCK.acquire_read(self.lock_timeout):
            raise HttpReqErrJson(503, "unable to take LOCK for reading after %s seconds" % self.lock_timeout)

        try:
            r =  self._do_request(request.get_method(), request.get_path(), None, request.get_headers())
            if r.text:
                return HttpResponseJson(r.status_code, r.json())

            return HttpResponseJson(r.status_code)
        except HttpReqErrJson, e:
            raise
        except Exception, e:
            LOG.exception("%r", e)
        finally:
            self.LOCK.release()


    ENDPOINT_POST_QSCHEMA = xys.load("""
    server_id?:   !!str
    endpoint?:    !!str
    """)

    ENDPOINT_POST_PSCHEMA = xys.load("""
    nameservers?:  [ !!str ]
    masters?:      [ !!str ]
    kind?:         !~~enum(native,master,slave)
    name?:         !!str
    soa_edit_api?: !~~enum(INCEPTION-INCREMENT,EPOCH,INCEPTION-EPOCH)
    """)

    def api_endpoint_post(self, request):
        params  = request.query_params()
        args    = request.payload_params()

        api_key = request.headers.get('x-api-key')
        if not api_key:
            raise HttpReqErrJson("missing api key")

        if not isinstance(params, dict):
            raise HttpReqErrJson(400, "invalid arguments type for query parameters")

        if not xys.validate(params, self.ENDPOINT_POST_QSCHEMA):
            raise HttpReqErrJson(415, "invalid arguments for command for query parameters")

        if not isinstance(args, dict):
            raise HttpReqErrJson(400, "invalid arguments type for payload parameters")

        if not xys.validate(args, self.ENDPOINT_POST_PSCHEMA):
            raise HttpReqErrJson(415, "invalid arguments for command payload parameters")

        if not self.LOCK.acquire_write(self.lock_timeout):
            raise HttpReqErrJson(503, "unable to take LOCK for reading after %s seconds" % self.lock_timeout)

        try:
            uids = []

            if params.get('endpoint') == 'zones':
                uids = self._push_apis_sync('create_hosted_zone', params, args)
            else:
                raise HttpReqErrJson(400, "invalid request")

            if not uids:
                raise HttpReqErrJson(500, "unable to retrieve uids")

            res = self._get_results(uids)
            if res['failed']:
                raise HttpReqErrJson(409, "failed to synchronize on dns provider. (errors: %r)" % res['failed'])

            r = self._do_request(request.get_method(), request.get_path(), args, request.get_headers())
            if r.text:
                return HttpResponseJson(r.status_code, r.json())

            return HttpResponseJson(r.status_code)
        except HttpReqErrJson, e:
            raise
        except Exception, e:
            LOG.exception("%r", e)
        finally:
            self.LOCK.release()


    ENDPOINT_PATCH_QSCHEMA = xys.load("""
    server_id:    !!str
    endpoint:     !!str
    id:           !!str
    """)

    ENDPOINT_PATCH_PSCHEMA = xys.load("""
    nameservers?:  [ !!str ]
    masters?:      [ !!str ]
    kind?:         !~~enum(native,master,slave)
    name?:         !!str
    soa_edit_api?: !~~enum(INCEPTION-INCREMENT,EPOCH,INCEPTION-EPOCH)
    rrsets?:       !~~seqlen(0,999)
      - records:
          - content:  !!str
            disabled: !~~isBool
        changetype:   'REPLACE'
        type:    !~~enum(A,AAAA,CNAME,RCNAME,MX,NS,TXT)
        name:    !!str
        ttl?:    !~~uint
    """)

    def api_endpoint_patch(self, request):
        params  = request.query_params()
        args    = request.payload_params()

        api_key = request.headers.get('x-api-key')
        if not api_key:
            raise HttpReqErrJson("missing api key")

        if not isinstance(params, dict):
            raise HttpReqErrJson(400, "invalid arguments type for query parameters")

        if not xys.validate(params, self.ENDPOINT_PATCH_QSCHEMA):
            raise HttpReqErrJson(415, "invalid arguments for command for query parameters")

        if not isinstance(args, dict):
            raise HttpReqErrJson(400, "invalid arguments type for payload parameters")

        if not xys.validate(args, self.ENDPOINT_PATCH_PSCHEMA):
            raise HttpReqErrJson(415, "invalid arguments for command payload parameters")

        if not self.LOCK.acquire_write(self.lock_timeout):
            raise HttpReqErrJson(503, "unable to take LOCK for reading after %s seconds" % self.lock_timeout)

        try:
            uids = []

            if params.get('endpoint') == 'zones':
                uids = self._push_apis_sync('change_rrsets', params, args)
            else:
                raise HttpReqErrJson(400, "invalid request")

            if not uids:
                raise HttpReqErrJson(500, "unable to retrieve uids")

            res = self._get_results(uids)
            if res['failed']:
                raise HttpReqErrJson(409, "failed to synchronize on dns provider. (errors: %r)" % res['failed'])

            r = self._do_request(request.get_method(), request.get_path(), args, request.get_headers())
            if r.text:
                return HttpResponseJson(r.status_code, r.json())

            return HttpResponseJson(r.status_code)
        except HttpReqErrJson, e:
            raise
        except Exception, e:
            LOG.exception("%r", e)
        finally:
            self.LOCK.release()

    ENDPOINT_DELETE_QSCHEMA = xys.load("""
    server_id:   !!str
    endpoint:    !!str
    id:          !!str
    """)

    def api_endpoint_delete(self, request):
        params  = request.query_params()

        api_key = request.headers.get('x-api-key')
        if not api_key:
            raise HttpReqErrJson("missing api key")

        if not isinstance(params, dict):
            raise HttpReqErrJson(400, "invalid arguments type for query parameters")

        if not xys.validate(params, self.ENDPOINT_DELETE_QSCHEMA):
            raise HttpReqErrJson(415, "invalid arguments for command for query parameters")

        if not self.LOCK.acquire_write(self.lock_timeout):
            raise HttpReqErrJson(503, "unable to take LOCK for reading after %s seconds" % self.lock_timeout)

        try:
            uids = []

            if params.get('endpoint') == 'zones':
                uids = self._push_apis_sync('delete_hosted_zone', params)
            else:
                raise HttpReqErrJson(400, "invalid request")

            if not uids:
                raise HttpReqErrJson(500, "unable to retrieve uids")

            res = self._get_results(uids)
            if res['failed']:
                raise HttpReqErrJson(409, "failed to synchronize on dns provider. (errors: %r)" % res['failed'])

            r = self._do_request(request.get_method(), request.get_path(), None, request.get_headers())
            if r.text:
                return HttpResponseJson(r.status_code, r.json())

            return HttpResponseJson(r.status_code)
        except HttpReqErrJson:
            raise
        except Exception, e:
            LOG.exception("%r", e)
        finally:
            self.LOCK.release()


    INDEX_SCHEMA    = xys.load("""
    server_id?: !!str
    endpoint?:  !!str
    method:     !!str
    params?:    !~~callback(pdns.validate_params)
    id:         !~~digit
    """)

    def index(self, request):
        args              = request.payload_params()
        api_key           = request.headers.get('x-api-key')

        if not isinstance(args, dict):
            raise HttpReqErrJson(400, "invalid arguments type")

        if not xys.validate(args, self.INDEX_SCHEMA):
            raise HttpReqErrJson(415, "invalid arguments for command")

        if not self.LOCK.acquire_write(self.lock_timeout):
            raise HttpReqErrJson(503, "unable to take LOCK for reading after %s seconds" % self.lock_timeout)

        try:
            method       = "_method_%s" % args['method'].replace('.', '_')
            self.ga_user = urllib.unquote(cookies.get('grafana_user').value)
            self.ga_sess = urllib.unquote(cookies.get('grafana_sess').value)
            self.ga_key  = "%s/%s" % (self.ga_user, self.ga_sess)

            for xid in KEYS_ID:
                self.ids["%ss" % xid] = []

            if 'params' in args:
                for xid in KEYS_ID:
                    xids = "%ss" % xid
                    if xids in args['params']:
                        self.ids[xids] = self._decrypt_ids(xid, args['params'][xids])

            if method != '_method_user_login':
                self.salt = self.adapter_redis.get_key("salt:%s" % self.ga_key)
                if not self.salt:
                    raise HttpReqErrJson(403, "forbidden")

                args = self._set_auth_arg(args)

            if method == '_method_user_login':
                r = getattr(self, method)(args, request)
            elif hasattr(self, method):
                r = getattr(self, method)(args, request)
            else:
                raise HttpReqErrJson(501, "unknown method: %r" % method)

            return self._build_http_response(r)
        except HttpReqErrJson:
            raise
        except Exception, e:
            LOG.exception("%r", e)
        finally:
            self.LOCK.release()


if __name__ != "__main__":
    def _start():
        MODULES.register(PDNSModule())
    _start()
